@@ -1,6 +1,7 @@
 package app_test
 
 import (
+	"bufio"
 	"context"
 	"crypto/tls"
 	"crypto/x509"
@@ -8,21 +9,24 @@ import (
 	"encoding/binary"
 	"fmt"
 	"io/ioutil"
+	"net"
+	"net/http"
+	"net/url"
 	"os"
 	"path/filepath"
 	"testing"
 	"time"
 
+	"github.com/canonical/go-dqlite"
 	"github.com/canonical/go-dqlite/app"
 	"github.com/canonical/go-dqlite/client"
-	"github.com/canonical/go-dqlite/internal/bindings"
 	"github.com/stretchr/testify/assert"
 	"github.com/stretchr/testify/require"
 )
 
 // Create a pristine bootstrap node with default value.
 func TestNew_PristineDefault(t *testing.T) {
-	_, cleanup := newApp(t)
+	_, cleanup := newApp(t, app.WithAddress("127.0.0.1:9000"))
 	defer cleanup()
 }
 
@@ -537,7 +541,7 @@ func TestRolesAdjustment_ReplaceVoter(t *testing.T) {
 		addr := fmt.Sprintf("127.0.0.1:900%d", i+1)
 		options := []app.Option{
 			app.WithAddress(addr),
-			app.WithRolesAdjustmentFrequency(500 * time.Millisecond),
+			app.WithRolesAdjustmentFrequency(2 * time.Second),
 		}
 		if i > 0 {
 			options = append(options, app.WithCluster([]string{"127.0.0.1:9001"}))
@@ -558,7 +562,7 @@ func TestRolesAdjustment_ReplaceVoter(t *testing.T) {
 	// A voter goes offline.
 	cleanups[2]()
 
-	time.Sleep(2 * time.Second)
+	time.Sleep(8 * time.Second)
 
 	cli, err := apps[0].Leader(context.Background())
 	require.NoError(t, err)
@@ -585,7 +589,7 @@ func TestRolesAdjustment_ReplaceVoterHonorFailureDomain(t *testing.T) {
 		addr := fmt.Sprintf("127.0.0.1:900%d", i+1)
 		options := []app.Option{
 			app.WithAddress(addr),
-			app.WithRolesAdjustmentFrequency(500 * time.Millisecond),
+			app.WithRolesAdjustmentFrequency(4 * time.Second),
 			app.WithFailureDomain(uint64(i % 3)),
 		}
 		if i > 0 {
@@ -609,7 +613,7 @@ func TestRolesAdjustment_ReplaceVoterHonorFailureDomain(t *testing.T) {
 	// A voter in failure domain 2 goes offline.
 	cleanups[2]()
 
-	time.Sleep(2 * time.Second)
+	time.Sleep(12 * time.Second)
 
 	cli, err := apps[0].Leader(context.Background())
 	require.NoError(t, err)
@@ -638,7 +642,7 @@ func TestRolesAdjustment_ReplaceVoterHonorWeight(t *testing.T) {
 		addr := fmt.Sprintf("127.0.0.1:900%d", i+1)
 		options := []app.Option{
 			app.WithAddress(addr),
-			app.WithRolesAdjustmentFrequency(500 * time.Millisecond),
+			app.WithRolesAdjustmentFrequency(4 * time.Second),
 		}
 		if i > 0 {
 			options = append(options, app.WithCluster([]string{"127.0.0.1:9001"}))
@@ -676,7 +680,7 @@ func TestRolesAdjustment_ReplaceVoterHonorWeight(t *testing.T) {
 	require.NoError(t, cli.Weight(context.Background(), uint64(10)))
 	defer cli.Close()
 
-	time.Sleep(2 * time.Second)
+	time.Sleep(12 * time.Second)
 
 	cli, err = apps[0].Leader(context.Background())
 	require.NoError(t, err)
@@ -705,7 +709,7 @@ func TestRolesAdjustment_CantReplaceVoter(t *testing.T) {
 		addr := fmt.Sprintf("127.0.0.1:900%d", i+1)
 		options := []app.Option{
 			app.WithAddress(addr),
-			app.WithRolesAdjustmentFrequency(500 * time.Millisecond),
+			app.WithRolesAdjustmentFrequency(4 * time.Second),
 		}
 		if i > 0 {
 			options = append(options, app.WithCluster([]string{"127.0.0.1:9001"}))
@@ -726,7 +730,7 @@ func TestRolesAdjustment_CantReplaceVoter(t *testing.T) {
 	cleanups[3]()
 	cleanups[2]()
 
-	time.Sleep(2 * time.Second)
+	time.Sleep(12 * time.Second)
 
 	cli, err := apps[0].Leader(context.Background())
 	require.NoError(t, err)
@@ -751,7 +755,7 @@ func TestRolesAdjustment_ReplaceStandBy(t *testing.T) {
 		addr := fmt.Sprintf("127.0.0.1:900%d", i+1)
 		options := []app.Option{
 			app.WithAddress(addr),
-			app.WithRolesAdjustmentFrequency(500 * time.Millisecond),
+			app.WithRolesAdjustmentFrequency(5 * time.Second),
 		}
 		if i > 0 {
 			options = append(options, app.WithCluster([]string{"127.0.0.1:9001"}))
@@ -775,7 +779,7 @@ func TestRolesAdjustment_ReplaceStandBy(t *testing.T) {
 	// A stand-by goes offline.
 	cleanups[4]()
 
-	time.Sleep(5 * time.Second)
+	time.Sleep(15 * time.Second)
 
 	cli, err := apps[0].Leader(context.Background())
 	require.NoError(t, err)
@@ -805,7 +809,7 @@ func TestRolesAdjustment_ReplaceStandByHonorFailureDomains(t *testing.T) {
 		addr := fmt.Sprintf("127.0.0.1:900%d", i+1)
 		options := []app.Option{
 			app.WithAddress(addr),
-			app.WithRolesAdjustmentFrequency(500 * time.Millisecond),
+			app.WithRolesAdjustmentFrequency(5 * time.Second),
 			app.WithFailureDomain(uint64(i % 3)),
 		}
 		if i > 0 {
@@ -832,7 +836,7 @@ func TestRolesAdjustment_ReplaceStandByHonorFailureDomains(t *testing.T) {
 	// A stand-by from failure domain 1 goes offline.
 	cleanups[4]()
 
-	time.Sleep(5 * time.Second)
+	time.Sleep(15 * time.Second)
 
 	cli, err := apps[0].Leader(context.Background())
 	require.NoError(t, err)
@@ -855,7 +859,7 @@ func TestRolesAdjustment_ReplaceStandByHonorFailureDomains(t *testing.T) {
 
 // Open a database on a fresh one-node cluster.
 func TestOpen(t *testing.T) {
-	app, cleanup := newApp(t)
+	app, cleanup := newApp(t, app.WithAddress("127.0.0.1:9000"))
 	defer cleanup()
 
 	db, err := app.Open(context.Background(), "test")
@@ -869,8 +873,9 @@ func TestOpen(t *testing.T) {
 // Test some setup options
 func TestOptions(t *testing.T) {
 	options := []app.Option{
+		app.WithAddress("127.0.0.1:9000"),
 		app.WithNetworkLatency(20 * time.Millisecond),
-		app.WithSnapshotParams(bindings.SnapshotParams{Threshold: 1024, Trailing: 1024}),
+		app.WithSnapshotParams(dqlite.SnapshotParams{Threshold: 1024, Trailing: 1024}),
 	}
 	app, cleanup := newApp(t, options...)
 	defer cleanup()
@@ -946,6 +951,82 @@ func newApp(t *testing.T, options ...app.Option) (*app.App, func()) {
 	return app, cleanup
 }
 
+// TestExternalConn creates a 3-member cluster using external http connections
+// and ensures the cluster is successfully created, and that the connection is
+// handled manually.
+func TestExternalConn(t *testing.T) {
+	externalAddr1 := "127.0.0.1:9191"
+	externalAddr2 := "127.0.0.1:9292"
+	externalAddr3 := "127.0.0.1:9393"
+	acceptCh1 := make(chan net.Conn)
+	acceptCh2 := make(chan net.Conn)
+	acceptCh3 := make(chan net.Conn)
+	hijackStatus := "101 Switching Protocols"
+
+	dialFunc := func(ctx context.Context, addr string) (net.Conn, error) {
+		conn, err := net.Dial("tcp", addr)
+		require.NoError(t, err)
+
+		request := &http.Request{}
+		request.URL, err = url.Parse("http://" + addr)
+		require.NoError(t, err)
+
+		require.NoError(t, request.Write(conn))
+		resp, err := http.ReadResponse(bufio.NewReader(conn), request)
+		require.NoError(t, err)
+		require.Equal(t, hijackStatus, resp.Status)
+
+		return conn, nil
+	}
+
+	newHandler := func(acceptCh chan net.Conn) http.HandlerFunc {
+		return func(w http.ResponseWriter, r *http.Request) {
+			hijacker, ok := w.(http.Hijacker)
+			require.True(t, ok)
+
+			conn, _, err := hijacker.Hijack()
+			require.NoError(t, err)
+
+			acceptCh <- conn
+		}
+	}
+
+	// Start up three listeners.
+	go http.ListenAndServe(externalAddr1, newHandler(acceptCh1))
+	go http.ListenAndServe(externalAddr2, newHandler(acceptCh2))
+	go http.ListenAndServe(externalAddr3, newHandler(acceptCh3))
+
+	app1, cleanup := newAppWithNoTLS(t, app.WithAddress(externalAddr1), app.WithExternalConn(dialFunc, acceptCh1))
+	defer cleanup()
+
+	app2, cleanup := newAppWithNoTLS(t, app.WithAddress(externalAddr2), app.WithExternalConn(dialFunc, acceptCh2), app.WithCluster([]string{externalAddr1}))
+	defer cleanup()
+
+	require.NoError(t, app2.Ready(context.Background()))
+
+	app3, cleanup := newAppWithNoTLS(t, app.WithAddress(externalAddr3), app.WithExternalConn(dialFunc, acceptCh3), app.WithCluster([]string{externalAddr1}))
+	defer cleanup()
+
+	require.NoError(t, app3.Ready(context.Background()))
+
+	// Get a client from the first node (likely the leader).
+	cli, err := app1.Leader(context.Background())
+	require.NoError(t, err)
+	defer cli.Close()
+
+	// Ensure entries exist for each cluster member.
+	cluster, err := cli.Cluster(context.Background())
+	require.NoError(t, err)
+	assert.Equal(t, externalAddr1, cluster[0].Address)
+	assert.Equal(t, externalAddr2, cluster[1].Address)
+	assert.Equal(t, externalAddr3, cluster[2].Address)
+
+	// Every cluster member should be a voter.
+	assert.Equal(t, client.Voter, cluster[0].Role)
+	assert.Equal(t, client.Voter, cluster[1].Role)
+	assert.Equal(t, client.Voter, cluster[2].Role)
+}
+
 func newAppWithDir(t *testing.T, dir string, options ...app.Option) (*app.App, func()) {
 	t.Helper()
 
@@ -970,6 +1051,31 @@ func newAppWithDir(t *testing.T, dir string, options ...app.Option) (*app.App, f
 	return app, cleanup
 }
 
+func newAppWithNoTLS(t *testing.T, options ...app.Option) (*app.App, func()) {
+	t.Helper()
+	dir, dirCleanup := newDir(t)
+
+	appIndex++
+
+	index := appIndex
+	log := func(l client.LogLevel, format string, a ...interface{}) {
+		format = fmt.Sprintf("%s - %d: %s: %s", time.Now().Format("15:04:01.000"), index, l.String(), format)
+		t.Logf(format, a...)
+	}
+
+	options = append(options, app.WithLogFunc(log))
+
+	app, err := app.New(dir, options...)
+	require.NoError(t, err)
+
+	cleanup := func() {
+		require.NoError(t, app.Close())
+		dirCleanup()
+	}
+
+	return app, cleanup
+}
+
 // Loads the test TLS certificates.
 func loadCert(t *testing.T) (tls.Certificate, *x509.CertPool) {
 	t.Helper()
